 /*
    MediumAccessControl.cc
    Last edited 21/09/2005 6:46pm by Luca De Nardis
*/
//i cout  vengono mostrati solo se si esegue in modalit run, non fast o express

#include <cmath>
#include <omnetpp.h>
#include "time.h"

#include <iomanip.h>
#include <vector.h>
#include "fstream.h"
#include "./Libs/PacketDescriptor.h"
#include "./Libs/FPL.h"
#include "./Libs/Connection.h"
#include "./Libs/connList.h"
#include "./Libs/MediumAccessControl.h"
using namespace std;



// ##########################################
// Method name    : 
// Description    :         /*
//* Friis free space equation:
//*
//*       Pt * Gt * Gr * (lambda^2)
//*   P = --------------------------
//*       (4 * pi * d)^2 * L
//
//* Pt    (4 * pi * d)^2 * L
//* -- = --------------------
//* Pr   Gt * Gr * (lambda^2)

// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double Friis(double Gt, double Gr, double lambda, double L, double d)
{


  double M = lambda/(4 * PI * d);
  return (1/Gt) * (1/Gr) * (1/M) * (1/M) * L;
}

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double Tarokh_pathLoss(double Xt, double Yt, double Xr, double Yr)
{
  //double L = ifp->getL();		// system loss
  //double lambda = ifp->getLambda();   // wavelength

	//	double Xt, Yt;		// loc of transmitter
	//double Xr, Yr;		// loc of receiver

	//	t->getNode()->getLoc(&Xt, &Yt, &Zt);
	//	r->getNode()->getLoc(&Xr, &Yr, &Zr);

	// Is antenna position relative to node position?
	/*
	  Xr += r->getAntenna()->getX();
	  Yr += r->getAntenna()->getY();
	  Zr += r->getAntenna()->getZ();
	  Xt += t->getAntenna()->getX();
	  Yt += t->getAntenna()->getY();
	  Zt += t->getAntenna()->getZ();
	*/
  double PL;
  double dX = Xr - Xt;
  double dY = Yr - Yt;
  double dZ = 0;

  double dist = sqrt(dX * dX + dY * dY + dZ * dZ);
  
  // get antenna gain
  double Gt = 1;
  double Gr = 1;
  double Pr = 0;
  double L=1;
  double n1_low_=-0.75;
  double n1_high_=0.75;
  double n2_low_=-2;
  double n2_high_=2;
  double n3_low_=-2;
  double n3_high_=2;
	
  // reference distance has ~ 47db
  double PL0 = Friis(Gt, Gr, lambda, L, dist0_);
	
  if (dist < dist0_)
  {
    // Free Space below reference distance
    PL = Friis(Gt, Gr, lambda, L, dist);
    if (PL < 1)
      PL=1;
  } else 
	{
    //double powerLoss_db = avg_db + ranVar->normal(0.0, std_db_);
    
    double average_powerLoss_db = 10.0 * average_pathlossExp_ * log10(dist/dist0_); //10 * mi_gamma * log_10 (d) in ref.
    
    double n1 = normal(0,1);
    while ( n1 < n1_low_ || n1 > n1_high_)
      n1 = normal(0,1);
    double n2 = normal(0,1);
    while ( n2 < n2_low_ || n2 > n2_high_)
      n2 = normal(0,1);
    double n3 = normal(0,1);
    while ( n3 < n3_low_ || n3 > n3_high_)
      n3 = normal(0,1);
    
    
    double random_powerLoss_db = 10 * n1 * pathlossExp_stdev_ * log10(dist/dist0_) + n2 * averageShadow_dev + n2 * n3 * Shadow_dev_stdev;
    
    // calculate the receiving power at dist
    
    PL = PL0*pow(10.0, (average_powerLoss_db + random_powerLoss_db)/10.0);
  }	
  return PL;
}

// ##########################################
// Method name    : optics_pathLoss
// Description    : pathloss for optical channel 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double optics_pathLoss(double xs, double ys, double xd, double yd, bool LOS)
{
double distance = sqrt(((xd - xs) * (xd - xs)) + ((yd - ys) * (yd - ys)));
double pathLossdB;
if(LOS)
{
	if (distance<=5)
	{
		pathLossdB=30+log10(pow((1+distance),40));
	}
    else
	{
		pathLossdB=53+distance*10/6;
	}
}	
else
{	
	pathLossdB=60+distance*10/6;
}
  double pl = pow(10.0, (pathLossdB / 10.0));
  return pl;
}


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double IEEE_802_15_4a_pathLoss(double xs, double ys, double xd, double yd)
{
    double distance = sqrt(((xd - xs) * (xd - xs)) + ((yd - ys) * (yd - ys)));
	shadow_term_802_15_4a=normal(0,1);
    double pathLossdB = PL0_802_15_4a[scenario] + 10 *pathloss_exp_802_15_4a[scenario]*log10(distance)+shadow_term_802_15_4a*shadow_stdv_802_15_4a[scenario];
    if (pathLossdB < 0.0)
	pathLossdB = 0.0;
	ev << "pathLossdB = " << pathLossdB << endl;
    double pl = pow(10.0, (pathLossdB / 10.0));
	return pl;
}

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double pathLoss(double xs, double ys, double xd, double yd)
{
    double distance = sqrt(((xd - xs) * (xd - xs)) + ((yd - ys) * (yd - ys))), pathLossdB = 47 + 17 * log10(distance), pl;
    if (pathLossdB < 0.0)
	pathLossdB = 0.0;
    pl = pow(10.0, (pathLossdB / 10.0));

//printf("\t\t\t>>> distanza %f        pl %e\n", distance, pl);
    return pl;
}

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double normalDistribution(double k, int n, double p)
{
    long double np = n * p;
    return 0.5L * (1.0L + erf((k - np) / sqrt(2 * np * (1.0L - p))));
}

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
int within(int lower, int upper, double prob, int n, double p)
{
    double lValue = normalDistribution((double)lower, n, p), uValue = normalDistribution((double)upper, n, p);
    if (prob < lValue)
	return -1;
    if (prob > uValue)
	return 1;
    return 0;
}


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
int inverseNormalDistribution(double prob, int n, double p)
{
//    int lower = (int)(n * p), interval = (int)(n * p * (1.0 - p)), result, count = 0;
    int lower = n / 2, interval = lower, result, count = 0;

    result = within(lower, lower + 1, prob, n, p);
    while (result != 0) {
	++count;
	if ((interval = interval / 2) == 0) {
//printf("CYCLE: %d\n", count);
	    return 0;
	}
	if (result == -1) {
	    if (lower == 0) {
//printf("CYCLE: %d\n", count);
		return 0;
	    }
	    lower = lower - interval;
	} else {
	    lower = lower + interval;
	    if (lower > n) {
//printf("CYCLE: %d\n", count);
		return n;
	    }
	}
	result = within(lower, lower + 1, prob, n, p);
    }
//printf("CYCLE: %d\n", count);
    return lower;
}

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double gammaln(double xx)
{
    double x, tmp, ser;
    static double cof[6] = {76.18009173, -86.50532033, 24.01409822, -1.231739516, 0.120858003e-02, -0.536382e-5};
    int j;
    
    x = xx - 1.0;
    tmp = x + 5.5;
    tmp -= (x + 0.5) * log (tmp);
    ser = 1.0;
    for (j = 0; j <= 5; ++j) {
	x += 1.0;
	ser += cof[j] / x;
    }
    return -tmp + log(2.50662827465 * ser);
}

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double factln(int n) 
{
//    static double a[1001];
    if (n <= 1) return 0.0;
//    if (n <= 1000) return a[n] ? a[n] : (a[n] = gammaln(n + 1.0));
    return gammaln(n + 1.0);
    return gammaln(n+1.0);
}


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double nchoosek(int n, int k) 
{
    return floor(0.5 + exp(factln(n) - factln(k) - factln(n-k)));
}


// ##########################################
// Method name    : Oghema(double A, double B)
// Description    : evaluation of the function "omega" used by the pulse collision model
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double Oghema(double A, double B)
{
    A = sqrt(A / 2);
    B = sqrt(B / 2);
    return (0.5 * erfc(A - B) + 0.5 * erfc(A + B) - erfc(A));
}


// ##########################################
// Method name    : Pcp
// Description    : factor used by the pulse collision model
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double Pcp(int Ni, int i, int Ns, double Pc0)
{
    double pcp = pow(Pc0, i) * pow(1.0 - Pc0, Ni * Ns - i);
    return nchoosek(Ni * Ns, i) * pcp;
}


// ##########################################
// Method name    : 
// Description    : ordina in modo decrescente 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
void sfalcosort(double *E, int length) {
    bool exchanged;
    int needle;
    double appo;

    do {
	exchanged = false;
	for (needle = 0; needle < length - 1; ++needle) {
//printf("\t E[%d] = %e < %e ?", needle, E[needle], E[needle + 1]);
	    if (E[needle] < E[needle + 1]) {
		appo = E[needle + 1];
		E[needle + 1] = E[needle];
		E[needle] = appo;
		exchanged = true;
//printf("\texchanged");
	    }
//printf("\n");
	}
    } while (exchanged);
}


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double Zmax(double *E, int Ni, int i)
{
    double Zmax = 0, appo;
    
    for (int j = 1; j <= Ni; ++j) {
//printf("\t\t coeff %d (%d) energy %e (i %d j %d Ni %d Es %e)\n", (int)ceil((i - j + 1) / (double)Ni), (i - j + 1) / Ni, sqrt(E[j-1]), i, j, Ni, E[j-1]);
	Zmax +=  sqrt(E[j-1]) * ceil((i - j + 1) / (double)Ni);
    }
//printf("\t\t\tZmax = %e (Zmax^2 %e)\n", Zmax, Zmax*Zmax);
    return Zmax;
}

//#define MINIMO(a,b)	((a < b) ? a : b)


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
void print_coord_update (double time, int ind)
{
  ofstream graph;
  graph.open ("Graphics.dat", ios::app);
  graph << '4' << endl;
  graph << time << " " << ind << " " << database[ind].x << " " << database[ind].y << endl;
  graph.close ();
}


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
void print_coordinate (int indice, int esecuzione)
{
  ofstream fp;
  char str[20];
  sprintf (str, "position%d.vec\0", esecuzione);
  fp.open (str, ios::app);
  fp << indice << "	" << database[indice].x << "	" << database[indice].y << endl;
  fp.close ();
}


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
void inizializza_file_position (int indice, int esecuzione)
{
  ofstream fp;
  char str[20];
  sprintf (str, "position%d.vec\0", esecuzione);
  fp.open (str, ios::app);
  fp << "vector " << indice << "  \"posizione\"" << "  \"nodo " << indice << "\"  1\n";
  fp.close ();
}



class MediumAccessControl:public cSimpleModule
{

  Module_Class_Members (MediumAccessControl, cSimpleModule, 4 * 8192)

    int numnodi;
  int TOsim;
  int esecuzione;
  bool mobilita;
  int vmax;
  double correlazionemobilita;
  double rangetx;
  double area_dim;
  double areadisplay;
  int mobilityupdate;
  double PRESERVINGINTERVAL;
  double mobilityinterval;
  double interarrivo;
  int TOmissingDATA;
  int TOmissingACKDATA;
  double maxPower;
  int DATA_discarded_noise;
  int LE_discarded_noise;
  int LC_discarded_noise;
  double AveragePEP, AverageBEP;
  double NbInterferentiNodo;
  int RecPackets;
  int interferringPackets[MAXDATASIZE];
  int DATAretransmitted;
  double retransmit_rate;
  bool BEB;
  double SlotDuration;
  int codedPHYSDUSize;
  int RSCorrectionCapability;
  int RSBlockLength;
  int MACPDUSize;
  int maxRetransmissionsPerPacket;

  struct statopacchetti
  {

    int DATAinviati;
    int DATAricevuti;
    int DATAmissedSync;
    int DATALost;
    int DATACollided;
    int DATADropped;
    int LEinviati;
    int LEricevuti;
    int LEmissedSync;

    int LCinviati;
    int LCricevuti;
    int LCmissedSync;

    int ACKDATAinviati;
    int ACKDATAricevuti;
    int ACKDATAmissedSync;
    
  }
  numeropacchetti;
  double packetdelay;
  double TDATA[NMAX];
  double TLC[NMAX];
  double TActive[NMAX];
  bool ActiveFor[NMAX];
  bool Active[NMAX];
  vector <int> TXQueue;
  int NPackets[NMAX];
  int indice;
  double xr;
  double yr;
  double vr;
  double dirr;
  bool finesim;
  double MissingDataSyncHeaderProb;
  double MissingCommonSyncHeaderProb;
  //potrei aggiungere un elemento
  struct statonode
  {
    int banda;
    double potenza;
    double ISI;
    int messaggiRRQ;
    int messaggiRRC;
  }
  nodo;

  struct locationinfo
  {
    bool valido;		//struttura che utilizzo per tavola routing
    double timestamp;
    double x;
    double y;
    int nexthop;
    double timenexthop;
  }
   *routingtable;

  int connID;
  int PDUNumber;
  FPL *pList;
  cStdDev delaystat;
 

// ##########################################
// Method name    : addErrors
// Description    : BER and PER evaluation
// Arguments      : 
// Called by      : 
// Returned value : int errorNew
// ##########################################
int addErrors(cMessage * msg)
  {
	
    int count;
	int txI = (int)msg->par("TXNodeID");
	int errorNew;
    //    double Ptx = 3.162278e+00, emui = 0.0, emuinorm, erx = 3.162278e+00 / pathLoss(nodeX[myI], nodeY[myI], nodeX[txI], nodeY[txI]),
    //     SIR, Rb = 1e6, sm2 = 1.9152e-10, SNR, SINR, BER, PER, packetLength = (double) msg->par("length") * 8.0L;
    //    double Ptx = 1.448313e-05, /* W */
    //    double Ptx = 5.36,
    //double Ts = 100e-09, /* average pulse repetition period */
    //Tm = 1e-9 * 0.9,	/* effective pulse duration */
    //Ns = 10.0,		/* number of pulses per bit */
    double N0 = noiseDensity,	/* monolateral power spectral density (W/Hz) */
	epsilon = Tm,		/* PPM shift */
	R0 = 0.0,
	Eu;
	double BER_before_MUI;
	//BER; // variabile per il BER, definita qui perche usata anche nell'ottica
		
		
    /* autocorrelation of the pulse shaper evaluated in epsilon */
	
	
    //energia utile ricevuta, il pathloss ottico credo che debba essere rivisto
	if(optics)
	{
		Eu = Ptx * Ts/optics_pathLoss(database[indice].x, database[indice].y, database[txI].x, database[txI].y,visibilityMatrix[txI][indice]);
	}
    else
	{
	
		if(IEEE_802_15_4_a)
		{
			//ev << "Considering 802_15_4a shadowing and delay spread " << endl;
			Eu = Ptx * Ts / IEEE_802_15_4a_pathLoss(database[indice].x, database[indice].y, database[txI].x, database[txI].y);	/* total energy per useful pulse*/
			ev<< "Node ["<<indice<<"]: Energy and power before multipath effect = " << Eu <<", "<< Eu/Ts << endl; 
			//Eu = Eu * PEAK_Matrix[txI][indice];		/*energy scaling modeling the effect of multipath*/
			ev<< "Node ["<<indice<<"]: Energy and power after multipath effect = " << Eu <<", "<< Eu/Ts << endl; 
		}
		else //Tarokh path loss
			
			Eu = Ptx * Ts / Tarokh_pathLoss(database[indice].x, database[indice].y, database[txI].x, database[txI].y);	/* energy per pulse with old pathloss model */
	}
    ev<< "Narrowband interference: " << NarrowBandInterf<< "Watt" <<endl;
	double EbN0 = Eu * Ns / (N0+NarrowBandInterf);  
	
		
	//	Pc0 = MINIMO(MINIMO(2.0 * Tm + epsilon, 4.0 * Tm), Ts) / Ts, // cambia qui introducendo tau_rms_max
	
	double *receivedImpulseEnergy = (double *)malloc(numnodi * sizeof(double));
	
	double BER = 0.5 * erfc(sqrt(0.5 * EbN0 * (1.0 - R0)));
	
	//debug purpose want to see the real impact of MUI on BER
	BER_before_MUI = BER;
    
	//cout << "N0= "<< N0 << "and narrowint= " << NarrowBandInterf << endl;
	
	//packetLength = (double) msg->par("length") * 8.0L; //Commented out by Luca, 21/03/05
    
    int Ni = 0;													//number of interfering users
	int M;														// number of interfering bits
	int packetLength =(int) msg->length()-PHYHeaderSize;		// length of the pkt
    double RMS_Max=0;
    
	ev<< "Node ["<<indice<<"]: packetLength = " << packetLength <<", Eb/N0 = "<< EbN0 << ", BER before interference = "<< BER << endl; 
	//cout << "Node ["<<indice<<"]: packetLength = " << packetLength <<", Eb/N0 = "<< EbN0 << ", BER before interference = "<< BER << endl; 
    //printf("Comm: %d --> %d @%f ================= EbN0 %e (Eu %e Ns %f N0 %e)\n", txI, indice, simTime(), EbN0, Eu, Ns, N0);
    //printf("\t -- Ptx %e, Ts %e, pl %e\n", Ptx, Ts, pathLoss(nodeX[indice], nodeY[indice], nodeX[txI], nodeY[txI]));
    
	for (count = 0; count < numnodi; ++count) 
	{
	
		if ((count == indice) || (count == txI)) //txI  il nodo di riferimento,  il ricevitore
		{
			receivedImpulseEnergy[count] = 0.0;
		} 
		else // considero ora tutti gli altri nodi del sistema
		{
			// ritardo di propagazione tra il nodo di riferimento (indice) e gli altri nodi diversi dal ricevitore (txI)
			double propagationDelay=sqrt(pow(database[count].x-database[indice].x,2)+pow(database[count].y-database[indice].y,2))/LIGHTSPEED;
			ev << "Propagation delay=  " << propagationDelay << endl;
			
			// calcolo dell'energia ricevuta per impulso da tutti i nodi interferenti. Questa energia  presente al ricevitore solo se il nodo count riesce a completare la trasmissione 
			// del pkt msg prima della fine della simulazione altrimenti l'energia interferente  nulla.
			if (endTransmission[count] + propagationDelay >= simTime()-(double)msg->length()/PACKETRATE)
			{
				
				ev<< "Node ["<<count<<"] interferring: endTransmission+propDelay = " << endTransmission[count]+propagationDelay <<endl;
				
				// incremento il numero degli interferenti
				++Ni; 

				//if (RMS_Matrix[count][indice] > RMS_Max)
				//RMS_Max = RMS_Matrix[count][indice];
				
				if(optics)
				{	
					//energia interferente dovuta al nodo count 
					receivedImpulseEnergy[count] = transmittedPower[count] * Ts / optics_pathLoss(database[indice].x, database[indice].y,database[count].x, database[count].y,visibilityMatrix[count][indice]); 
					
					// cout<<receivedImpulseEnergy[count]<<endl;
				}
				else
				{
					if(IEEE_802_15_4_a)
					{
						//ev << "Uso pathloss 802_15_4a" << endl;
			
						receivedImpulseEnergy[count] = transmittedPower[count] * Ts / IEEE_802_15_4a_pathLoss(database[indice].x, database[indice].y,database[count].x, database[count].y);
						ev<< "Node ["<<indice<<"]: Energy of interferer " << count << " before multipath effect = " << receivedImpulseEnergy[count] << endl;
						//ev << "Node ["<<indice<<"]: RMS_Matrix[" <<count <<"]["<<indice<<"] = "<<RMS_Matrix[count][indice] << endl;
						//receivedImpulseEnergy[count] = receivedImpulseEnergy[count] *Tm / RMS_Matrix[count][indice]; /* Tiene conto del delay spread */
						ev<< "Node ["<<indice<<"]: Energy of interferer " << count << " after multipath effect = " << receivedImpulseEnergy[count] << endl; 
					}
					
					else
						
						receivedImpulseEnergy[count] = transmittedPower[count] * Ts / Tarokh_pathLoss(database[indice].x, database[indice].y,database[count].x, database[count].y);
				}
			}
			else // non fa in tempo a trasmettere il msg, quindi impongo che l'energia ricevuta sia nulla
				receivedImpulseEnergy[count] = 0.0;
		
		}//end of else
	}//end of "for (count = 0; count < numnodi; ++count)"
    
	// maximum number of possible collisions during a bit period (Ns*Ts), used by pulse collision model
	M = Ni * (int) Ns; 
    //printf("\n-- Ni %d  Ns %d  M %d (BER = %e) --- %p -- E[0] %e\n", Ni, (int)Ns, M, BER, receivedImpulseEnergy, receivedImpulseEnergy[0]);
    
	//updating variable for statistics
	NbInterferentiNodo+=Ni;
	//cout << NbInterferentiNodo[indice] << endl;
	
	if (!optics)
	{
		sfalcosort(receivedImpulseEnergy, numnodi); // decrescente
    
		for(count=0; count < numnodi; ++count) 
		{
			ev << "receivedImpulseEnergy[" << count<< "]: "<< receivedImpulseEnergy[count] << endl;
		}
		//double Pc0 = min(min(2*Tm + epsilon, 4*Tm),Ts) / Ts;
		double Pc0 = min(min(2*Tm + epsilon, 4*Tm),Ts) / Ts;
		
		//applying the pulse collision model for BER evaluation
		for (count = 0; count < M; ++count) 
		{
			//    for (count = Ni; count < M; ++count) {
			double primo, zmax, oghema;
			primo = 0.5 * pow(Pcp(Ni, count, (int) Ns, Pc0),2);
			zmax = Zmax(receivedImpulseEnergy, Ni, count);
			oghema = Oghema(EbN0 * (1 - R0), pow(zmax, 2) / (Ns * N0 * (1 - R0)));
			BER += primo * oghema;
		}
		
		/*int numBlocks = packetLength/RSBlockLength;
		int errorRS=0;
		for (int i = 0;i<numBlocks;i++)
		{
			int errorsNumberBlock = inverseNormalDistribution((double)uniform(0, 1), (int) RSBlockLength, BER);
			ev << "Errors with Salvatore's formula in block "<< i << " = " << errorsNumberBlock << endl;
			if (errorsNumberBlock >= RSCorrectionCapability)
			{
				errorRS=1;
				break;
			}
		}	
		//printf("interf. %d -- BER %e\t numero errori %d\n", Ni, BER, errorsNumber);*/
	
	}// end of (!optics)
	if(optics) //if (optics == true)
	{
		//cout << FreqLED <<"is ? " << pow(3.529,14) << endl;
		
		sfalcosort(receivedImpulseEnergy, numnodi); // ordinamento decrescente (dei nodi rispetto all'energia ricevuta per impulso???)
		double numeratore1=Eu;
		double numeratore2=0;
		double denominatore=FreqLED* 6.6260*pow(10.0,-34); //freq relativa al led passato come parametro * Plank const ( l'energia per fotone)
		
		for (int i=0; i<Ni; i++)
	    {
			numeratore1=numeratore1+Tm/Ts*0.5*receivedImpulseEnergy[i];
			//cout << "num1 = " << numeratore1 << endl;
			numeratore2+=0.5*Tm/Ts*receivedImpulseEnergy[i];
			//cout << "num2 = " << numeratore2 << endl;
	    }
		
		//questo BER non tiene conto affatto del rumore termico..
		BER=0.5*exp(-numeratore1/denominatore)+0.5*(1-exp(-numeratore2/denominatore));
	  
		  //cout << "tutto torna" << endl;
		  //BER=Ni*Tm/Ts; // simple BER evaluation
		  //if(Ni>=1)
		  //{
		  //cout << "Eu = " << Eu << endl;
		  //  cout << "num-Eu = " << numeratore1-Eu << endl;
		  //cout << "ni = " << Ni << endl;
		  //  cout << "ber = " << BER << endl;
		  //}
	} //end of else (optics == true)
	if(DWO_OOC)
	{
		//receiver responsivity
		//double Rho = Eta*q/FreqLED* pow(6.6260,-34);
		
		//evaluate gamma
		
		//double Gamma = Rho*Ptx
//		cout << "OocLenght" << OocLenght<< endl;
//		cout << "OocWeight" << OocWeight<< endl;
//		cout << "Rho" << Rho<< endl;
//		cout << "Eta" << Eta<< endl;
//		cout << "CeilingHeight" << CeilingHeight<< endl;
//		cout << "OocRecThreshold" << OocRecThreshold<< endl;
	
		//evaluation of the maximum number of users
		int MaxNbUsers=ceil((OocLenght-1)/(OocWeight*(OocWeight-1)));
		//cout << MaxNbUsers<< endl;
		//evaluation of the interfering contributions
		
		//P0 and P1 evaluation
		
		//Pm0 and Pm1 evaluation
		
		
		
		//BER evaluation
	}
 
    
    double packetErrorProbability = 1 - pow(1 - BER, packetLength);
    
    ev<< "BER after = "<< BER<< ", packetErrorProbability = " << packetErrorProbability << endl;

    
    if (uniform(0, 1) < packetErrorProbability)
      errorNew=1;
    else
      errorNew=0;
	//debug purpose
	//AverageBEP += BER_before_MUI;
	//updating statistics for PEP end BEP
	AverageBEP += BER;
	//cout << "average bep" << AverageBEP << endl;
	AveragePEP += packetErrorProbability;
	RecPackets++;
    
	//ev << "Error = "<< errorNew <<"; ErrorRS = " << errorRS << endl;
    free(receivedImpulseEnergy);
    
    //return errorRS;
    return errorNew;
	
  } //end of "addErrors(cMessage * msg)"
  
// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
  double BCP(int Nu)
  {
    return(exp((-2*Nu*PULSELENGTH*PACKETRATE)));
  }


// ##########################################
// Method name    : evaluateErrorProbability
// Description    : this function evaluates the PEP (packet error probability) and send it back to caller function
// note			  :
// Arguments      : 
// Called by      : validPacketCheck that is called only if PCMUIModel is true, that is if  PCENHANCED==false
// Returned value : double
// ##########################################
//double evaluateErrorProbability (int nodeID, double txPower, int txNodeID, double intPower, int packetSize, int packetKind)
//{
//	
//	ofstream intLE, intLC, intDATA;
//	double packetErrorProbability;
//	double distTxRx = sqrt (pow (database[nodeID].x - database[txNodeID].x, 2) + pow (database[nodeID].y - database[txNodeID].y, 2));
//	
//	if(!PCMUIModel)
//	{
//		double SNR = (txPower / pow (distTxRx, 2)) / (intPower + noisePower);
//		//ev << "Node [" << indice << "]: Potenza ricevuta="<< txPower/pow(distTxRx,2)<<", Potenza interferente=" << intPower << ", Potenza di rumore = " << noisePower <<  ", SNR ottenuto = " << SNR << "; soglia con margine = " << sogliaSNR * MUI_margin << endl;
//
//		//    if(SNR < sogliaSNR * MUI_margin)
//		// ev << "Node [" << indice << "]: Error in received power?" << endl;
//		
//		//Parte temporanea da sostituire con la funzione di probabilita'
//		double y = sqrt (SNR / 2);
//		double bitErrorProbability = 0.5 * erfc(y);
//		packetErrorProbability = 1 - pow(1 - bitErrorProbability, packetSize);
//	}
//	else
//	{
//		double packetCorrectProbability = 1;
//		for(int j = 0; j < packetSize; j++)
//		{
//			//ev << "Node [" << indice << "]: interferringPackets for bit i =" << interferringPackets[j] << endl;
//			packetCorrectProbability *= BCP(interferringPackets[j]);
//			//ev << "Node [" << indice << "]: BCP for bit i =" << BCP(interferringPackets[j]) << endl;
//		}
//
//		//double bitErrorProbability = 0.5 * erfc (y);
//		//double packetErrorProbability = 1 - pow (1 - bitErrorProbability, packetSize);
//		packetErrorProbability = 1 - packetCorrectProbability;
//		//ev << "Node [" << indice << "]: PEP =" << packetErrorProbability << endl;
//	}
//	if(packetKind == DATA)
//	{
//		//AverageBEP += bitErrorProbability;
//		AveragePEP += packetErrorProbability;
//		RecPackets++;
//	}
//	if (DEBUG)
//	{
//		//if(intPower > 0)
//		switch(packetKind)
//		{
//			case LE:
//			{
//				intLE.open("Interf_LE.txt", ios::app);
//				intLE << setprecision(10) << simTime() << " " << setiosflags(ios::scientific) << txPower/pow(distTxRx,2) << " " << intPower << " " << (txPower/pow(distTxRx,2))/intPower << " " << packetErrorProbability << endl;
//				intLE.close();
//			}
//				break;
//			case LC:
//			{
//				intLC.open("Interf_LC.txt", ios::app);
//				intLC << setprecision(10) << simTime() << " " << setiosflags(ios::scientific) << txPower/pow(distTxRx,2) << " " << intPower << " " << (txPower/pow(distTxRx,2))/intPower << " " << packetErrorProbability << endl;
//				intLC.close();
//			}
//				break;
//			case DATA:
//			{
//				intDATA.open("Interf_DATA.txt", ios::app);
//				intDATA << setprecision(10) << simTime() << " " << setiosflags(ios::scientific) << txPower/pow(distTxRx,2) << " " << intPower << " " << (txPower/pow(distTxRx,2))/intPower << " " << packetErrorProbability << endl;
//				intDATA.close();
//			}
//				break;
//		}
//
//	}
//	return (packetErrorProbability);
//}


// ##########################################
// Method name    : validPacketCheck
// Description    : this function first calls evaluateErrorProbability() to know the value of PEP. When PEP is known, 
//				  : a packet check can be made using a random uniform ditributed variable
// Arguments      : 
// Called by      : checkCorrectReceiving ONLY if the pulse collision model is NOT being used
// Returned value : bool
// ##########################################	
//bool validPacketCheck (int tipoPacchetto, int nodeID, int PDUNumber, double rxTime, int packetSize, double txPower, int txNodeID)
//{
//	//breakpoint("Sono in validPacketCheck");
//	double packetDuration= (double) packetSize / PACKETRATE;
//	double errorProbability;
//	
//	switch (tipoPacchetto)
//	{
//		case DATA:
//		{
//			if (!PCMUIModel) //it will never enter here, because validPacketCheck is called only if PCMUIModel is activated, isn't?
//			{
//
//				double intPower = packList->evaluateInterferencePower (nodeID, PDUNumber, packetDuration, rxTime, database[nodeID].x, database[nodeID].y, txNodeID, PACKETRATE);
//				errorProbability = evaluateErrorProbability (nodeID, txPower, txNodeID, intPower, packetSize, DATA);
//				ev << "Node [" << indice << "]: SGA model, errorProbability =" << errorProbability << endl; 
//			}
//			else
//			{
//				packList->evaluateCollidingPackets(nodeID, PDUNumber, packetSize, rxTime, database[nodeID].x, database[nodeID].y, PACKETRATE, txNodeID, interferringPackets);
//				//double errorProbability = evaluateErrorProbability (nodeID, txPower, txNodeID, intPower, packetSize, DATA);
//				errorProbability = evaluateErrorProbability (nodeID, txPower, txNodeID, 0, packetSize, DATA);
//				ev << "Node [" << indice << "]: PC model, errorProbability =" << errorProbability << endl; 
//			}
//			//random variable uniform distribution
//			double value = genk_uniform (1, 0, 1);
//			
//			if (value <= errorProbability)
//			{
//				DATA_discarded_noise ++;
//				
//				return (false);
//			}
//			else
//				return (true);
//		}
//			break;   
//		
//		case ACKDATA:
//		{
//			return (true);
//		}
//			break;
//
//		case LC:
//		{
//			double intPower = packList->evaluateInterferencePower (nodeID, PDUNumber, packetDuration, rxTime, database[nodeID].x, database[nodeID].y, txNodeID, PACKETRATE);
//			double errorProbability = evaluateErrorProbability (nodeID, txPower, txNodeID, intPower,
//						  packetSize, LC);
//			double value = genk_uniform (1, 0, 1);
//		if (value <= errorProbability)
//		{
//			LC_discarded_noise ++;
//			return (false);
//		}
//		else
//			return (true);
//		}
//			break;
//
//		case LE:
//		{
//			double intPower = packList->evaluateInterferencePower (nodeID, PDUNumber, packetDuration, rxTime, database[nodeID].x, database[nodeID].y, txNodeID, PACKETRATE);
//
//			double errorProbability = evaluateErrorProbability (nodeID, txPower, txNodeID, intPower, packetSize, LE);
//			double value = genk_uniform (1, 0, 1);
//		if (value <= errorProbability)
//		{
//			LE_discarded_noise ++;
//			return (false);
//		}
//		else
//			return (true);
//		}
//			break;
//
//		case DISTANCE_REQUEST:
//		{
//			return (true);
//		}
//			break;
//
//		case DISTANCE_REPLY:
//		{
//			return (true);
//		}
//			break;
//	}//end of switch
//	
//	return (true);
//}


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
  double evaluateTxPower (int destinationID)
  {
    double power;
    if (destinationID >= 0 && NeighbourMatrix[indice][destinationID].distanceValidity)
      {
	power = (NeighbourMatrix[indice][destinationID].distance <= rangetx) ? STDPOWER * pow (NeighbourMatrix[indice][destinationID].distance, 2) : maxPower;	
	ev << "Node [" << indice <<"] sets transmit power to " << power << " with maxPower = " << maxPower << endl;
      }
    else			//trasmissione broadcast a potenza fissa e massima
      {
	power = maxPower;
	ev << "Node [" << indice <<"] sets transmit power to maxPower = " << maxPower << endl;
      }
    return (power);
  }

// ##########################################
// Method name    : PacketEnergyEvaluation
// Description    : this function returns the energy consumption due to packet 
//					transmission or packet reception (according to the bool tx 
//					value), for the moment this is valid for UWB systems with about 1Mbps data rate.
//					This was coded referring to a paper that was not specific of UWB
// Arguments      : 
// Called by      : 
// Returned value : double UsedPacketEnergy
// ##########################################

double PacketEnergyEvaluation(double d, int packetSize, bool tx) 
{
	double UsedPacketEnergy;
	if (!optics || !DWO_OOC)
	{
		double alphaprop= 2; //propagation coefficient freespace = 2
		
		double E_start = 2.76e-5;
		double E_Tx_bit_fixed = 3.25e-7 * ((double) PACKETRATE/1e6);
		double E_Tx_bit_prop = 1.25e-12 * ((double) PACKETRATE/1e6);
		double E_Rx_bit =1.13e-7+2.79e-7 * ((double) PACKETRATE/1e6);
		
		if (tx)			//transmitting node
			UsedPacketEnergy = E_start + packetSize * (E_Tx_bit_fixed +  E_Tx_bit_prop * (pow(d, alphaprop)));
			
			
		else			//receiving node
			UsedPacketEnergy =  E_start + packetSize * E_Rx_bit;
		//ev << "Node["<<NodeIndex<<"] used " << UsedPacketEnergy << " J for " << (tx ? "tx":"rx") << " the packet"<< endl; 
    }
	else
	{
		UsedPacketEnergy=0;//cout << "qualcosa non va nella funzione di calcolo dell'energia UWB"<< endl;
	}
	
	return (UsedPacketEnergy);
	
}//end of PacketEnergyEvaluation()


// ##########################################
// Method name    : 
// Description    : 02/08/2002 Inserito da Luca: funzione che genera le MAC PDU 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
   
  void trafficGen (int *PDUNumber, int *receiverID)  
  {
    *PDUNumber = genk_intuniform(1,1,MAXPDU);
    *receiverID = indice;
    while((*receiverID) == indice)
      {
	    *receiverID = genk_intuniform(1,0,numnodi-1);
      }
  }
// ##########################################
// Method name    : 
// Description    : 02/08/2002 Inserito da Luca: funzione che effettua lo scheduling di un pacchetto
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
    
  bool packetScheduling (int packetSize, int connID, int RXNodeID, int *PDUNumber, int packetKind, double actualTime, cMessage * oldmsg)
  {

    double *txTime = new double;

    int nodeID = indice;
    double processingTime = processingBitTime * packetSize;
    ev << "Node [" << indice <<"] expectedTxTime: " << actualTime + processingTime << endl;
    double packetDuration = (double) (packetSize + PHYHeaderSize) / PACKETRATE;
    double SlotDelay = ceil(simTime()/SlotDuration)*SlotDuration-simTime();
    ev << "Node [" << indice <<"] SlotDelay: " << SlotDelay << endl;
    if (pList->insertInFPL(nodeID, *PDUNumber, RXNodeID, 0, actualTime + processingTime, packetKind, packetDuration, txTime, false, SlottedAloha, SlotDelay, SlotDuration))
      {
	ev << "Node [" << indice <<"] effectiveTxTime: " << *txTime << endl;
	if((DEBUG) && (indice == 2))
	  pList->printList();
	cMessage *msg = new cMessage;
	msg->setKind (SENDPACKET);
	msg->addPar ("packetKind") = packetKind;
	msg->addPar ("TXNodeID") = indice;
	msg->addPar ("PDUNumber") = *PDUNumber;
	msg->addPar ("RXNodeID") = RXNodeID;
	switch (packetKind)	//Aggiungiamo al messaggio le informazioni ed effettuiamo altre operazioni dipendenti dal suo tipo...
	  {
	  case ACKDATA:
	    {
	      msg->setLength(ACKDATASIZE + PHYHeaderSize);
	      msg->setName ("SENDPACKET-ACKDATA");
	    }
	    break;
	  case DATA:
	    {
	      msg->setLength(codedPHYSDUSize + PHYHeaderSize);
	      msg->setName ("SENDPACKET-DATA");
	      msg->addPar("NPackets") = NPackets[RXNodeID];
	      msg->addPar("TXTime")=oldmsg->par("TXTime"); //added
	      msg->addPar("TXAttempt")=oldmsg->par("TXAttempt");//added to avoid instability
	    }
	    break;
	  case LC:
	    {
	      msg->setLength(LCSIZE + PHYHeaderSize);
	      msg->setName ("SENDPACKET-LC");
	    }
	    break;
	  case LE:
	    {
	      msg->setLength(LESIZE + PHYHeaderSize);
	      msg->setName ("SENDPACKET-LE");
	      msg->addPar("TH-flag") = true;      
	    }
	    break;
	  case DISTANCE_REQUEST:
	    {
	      msg->setName ("SENDPACKET-DISTANCE_REQ");
	      msg->setLength (oldmsg->length () + PHYHeaderSize);
	      msg->addPar ("destinationNode") = oldmsg->par ("destinationNode");
	    }
	    break;
	  case DISTANCE_REPLY:
	    {
	      msg->addPar ("PDUNumber") = *PDUNumber;
	      msg->setName ("SENDPACKET-DISTANCE_REP");
	      msg->setLength (oldmsg->length () + PHYHeaderSize);
	      msg->addPar ("destinationNode") = oldmsg->par ("destinationNode");
	    }
	    break;
	  }
	if (*txTime < simTime ())
	  {
	    ev << "txTime= " << *txTime << " simTime= " << simTime () << " Differenza= " << simTime () - (*txTime) << endl;
	    //breakpoint ("Errore, pacchetto schedulato nel passato");
	  }
	scheduleAt (*txTime, msg);
	delete txTime;
	(*PDUNumber)++;
	return (true);
      }
    delete txTime;
    return (false);
  }


// ##########################################
// Method name    : checkCorrectReceiving
// Description    : /04/08/2002 Inserito da Luca: funzione che richiede al CM il 
//				  : controllo su un pacchetto ricevuto, l'out  true se il pacchetto  stato ben ricevuto
// Arguments      : 
// Called by      : activity()
// Returned value : bool
// ##########################################
  bool checkCorrectReceiving (int type, int packetSize, int PDUNumber, double rxTime, double txPower, int txNodeID, cMessage * recpkt)
  {
	  bool packetStatus;
	  //FEC_LIMIT is defined == 0
	  //test if the pulse collision model is being used
	  if(PC_ENHANCED)
		  
		  packetStatus = (addErrors(recpkt) > FEC_LIMIT)? 0 : 1;
	  
//	  else
//		  packetStatus = validPacketCheck (type, indice, PDUNumber, rxTime, packetSize, txPower, txNodeID);
	  
	  return packetStatus;
  }

// ##########################################
// Method name    : 
// Description    : modifica stato nodo RISERVANDO sulla base risorse richieste dalla connessione
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
  void RiservaRisorse (int bitRate)	
  {
    nodo.banda -= bitRate;
  }

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
  void liberaRisorse (int bitRate)	//modifica stato nodo LIBERANDO sulla base risorse richieste dalla connessione 
  {
    nodo.banda += bitRate;
  }
  /*  void Broadcast (cMessage * msg, double packetDuration)
  {
    ev << "Entro in broadcast\n";
    char str[6];
    int hopvec[NMAX];
    int Hops = 0 ;
    for (int i = 0; i < numnodi; i++)
      {
	if(msg->kind() == RRQ)
	  {
	    Hops = (int) msg->par("hops");
	    for (int j = 0; j <= Hops ; j++)
	      {
		sprintf(str, "hop%d",j);
		hopvec[j] = (int) msg->par(str);
	      }
	  }
	if (networkconnectivity[indice][i] && (indice != i)&&(Neighbour_not_in_path((int) msg->kind(),i,hopvec,Hops)))
	  {
	    ev << "Nodo " << indice << ": invio pacchetto a nodo " << i << endl;
	    cMessage *copy = (cMessage *) msg->dup ();
	    double propagationDelay = sqrt (pow (database[indice].x - database[i].x, 2) + pow (database[indice].y - database[i].y, 2)) / LIGHTSPEED;
	    sendDelayed (copy, propagationDelay + packetDuration, "outnode", i);
	    nodo.potenza -= TXmsg;
	  }
      }				// delete msg deve essere eseguito dopo la chiamata alla funzione
  }
  */

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
  double Speed (int percentage)
  {
    return (vmax);
    //DA COMPLETARE
  }


// ##########################################
// Method name    : evaluateGeneratedMUIcontribution
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
double evaluateGeneratedMUIcontribution(int nextHop)
{

	double interferenceContribution = 0;
	double dmin = INFINITY;
	int neighbours =0;
	for (int i = 0; i < numnodi; i++)
	{
		if (NeighbourMatrix[indice][i].distance != -1 && i != nextHop)
		{
			if (NeighbourMatrix[indice][i].distance < dmin)
				dmin = NeighbourMatrix[indice][i].distance;
			
			neighbours++;
		}
	}

	for (int i = 0; i < numnodi; i++)
	{
		if (NeighbourMatrix[indice][i].distance != -1 && i != nextHop)
			interferenceContribution += pow (1 - dmin / NeighbourMatrix[indice][i].distance, 2);
	}

	return((neighbours > 0) ? interferenceContribution/neighbours : 0);
}//end of "evaluateGeneratedMUIcontribution"

// ##########################################
// Method name    : serveTXQueue
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
void serveTXQueue(cMessage *msg, int RXNodeID = -1)
{
//breakpoint("In TXQueue");
int i = 0;
vector <int>::iterator iter =TXQueue.begin();
ev << "Node [" << indice << "] in serveTXQueue" << endl;
ev << "Node [" << indice << "]: PDUs in TXQueue = " << TXQueue.size() << endl;
while(iter != TXQueue.end())
  {
//breakpoint("TXQueue: in the while");
if(Active[*iter]||(((double) msg->par("TXTime")<simTime())&&(*iter == RXNodeID)))//added the condition after the OR
  {
	NPackets[*iter]--;
	if(networkconnectivity[indice][*iter])
	  packetScheduling (codedPHYSDUSize, 0, *iter, &PDUNumber, DATA, simTime (), msg);
	TXQueue.erase(iter);
  }
else
  {
	if(TXQueue.size()>0)
	  {
	if(TLC[*iter] < simTime())
	  TXQueue.erase(iter);
	else
	  {
		ev << "Node [" << indice << "] scheduling LE packet" << endl;
		if(*iter == RXNodeID)
		  {
		if(networkconnectivity[indice][*iter])
		  packetScheduling (LESIZE, 0, *iter, &PDUNumber, LE, simTime (), msg);
		ev << *iter << endl;
		iter = TXQueue.end();
		  }
		if(iter!=TXQueue.end())
		  ++iter;
	  }
	  }
  }
  }
//    breakpoint("Out of TXQueue");
}


  virtual void initialize ();
  virtual void activity ();
  virtual void finish ();

};

Define_Module (MediumAccessControl);


// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
void MediumAccessControl::initialize ()
{
  cPar & x = parentModule ()->par ("x");	//passaggio parametri per variabile (&)
  cPar & y = parentModule ()->par ("y");	//assegnando variabili x,y,dir,v modifico parametri moduli relativi
  cPar & px = parentModule ()->par ("px");
  cPar & py = parentModule ()->par ("py");
  cPar & dir = parentModule ()->par ("direction");
  cPar & v = parentModule ()->par ("speed");
  //cPar & NodeEnergy = parentModule ()->par ("NodeEnergy");
  
  InitialNodeEnergy = parentModule ()->parentModule ()->par ("InitialNodeEnergy");
  BEB = parentModule ()->parentModule ()->par ("BEB");
  numnodi = parentModule ()->par ("numnodi");	//sono valori costanti durante esecuzione
  esecuzione = parentModule ()->parentModule ()->par ("esecuzione");	//eseguo copia  parametri moduli relativi nelle variabili mv,vv....
  areadisplay = parentModule ()->parentModule ()->par ("areadisplay");
  TOsim = parentModule ()->parentModule ()->par ("TOsim");
  interarrivo = parentModule ()->parentModule ()->par ("interarrivo");
  
  //optical variables
  FreqLED = parentModule ()->parentModule ()->par ("FreqLED");
  OocLenght = parentModule ()->parentModule ()->par ("OocLenght");
  OocWeight = parentModule ()->parentModule ()->par ("OocWeight");
  
  Eta = parentModule ()->parentModule ()->par ("Eta");
  CeilingHeight = parentModule ()->parentModule ()->par ("CeilingHeight");
  OocRecThreshold = parentModule ()->parentModule ()->par ("OocRecThreshold");
  //end optical variables
  
  TOmissingDATA = parentModule ()->parentModule ()->par ("TOmissingDATA");
  TOmissingACKDATA = TOmissingDATA;
  mobilita = parentModule ()->parentModule ()->par ("mobilita");
  vmax = parentModule ()->parentModule ()->par ("vmax");
  mobilityinterval = parentModule ()->parentModule ()->par ("mobilityinterval");
  correlazionemobilita = parentModule ()->parentModule ()->par ("correlazionemobilita");
  area_dim = parentModule ()->parentModule ()->par ("area_dim");
  scenario = parentModule ()->parentModule ()->par ("scenario");
  optics = parentModule ()->parentModule ()->par ("optics");
  DWO_OOC = parentModule ()->parentModule ()->par ("DWO_OOC");
  rangetx = parentModule ()->parentModule ()->par ("rangetx");
  sogliaSNR = parentModule ()->parentModule ()->par ("SNRThreshold");
  noiseDensity = (double) parentModule ()->parentModule ()->par ("NoiseDensity") * (double)  parentModule ()->parentModule ()->par ("NoiseFactor");
  Ptx = parentModule ()->parentModule ()->par ("Ptx");
  MUI_margin = parentModule ()->parentModule ()->par ("MUIMargin");
  PCMUIModel = parentModule ()->parentModule ()->par ("PCMUIModel");
  PC_ENHANCED = parentModule ()->parentModule ()->par ("PC_ENHANCED");
  STDPOWER = noisePower * sogliaSNR * MUI_margin;
  SlottedAloha =  parentModule ()->parentModule ()->par ("SlottedAloha");
  IEEE_802_15_4_a = parentModule ()->parentModule ()->par ("IEEE_802_15_4_a");
  NarrowBandInterf = parentModule ()->parentModule ()->par ("NarrowBandInterf");
  average_pathlossExp_= parentModule ()->parentModule ()->par ("average_pathlossExp_");
  pathlossExp_stdev_= parentModule ()->parentModule ()->par ("pathlossExp_stdev_");
  averageShadow_dev= parentModule ()->parentModule ()->par ("averageShadow_dev");
  Shadow_dev_stdev= parentModule ()->parentModule ()->par ("Shadow_dev_stdev");
  lambda =  parentModule ()->parentModule ()->par ("lambda");
  dist0_ =  parentModule ()->parentModule ()->par ("dist0");
  Ts = parentModule ()->parentModule ()->par ("Ts");
  Tm = parentModule ()->parentModule ()->par ("Tm");
  Ns = parentModule ()->parentModule ()->par ("Ns");
  PACKETRATE = 1/(Ts*Ns);
  noisePower = noiseDensity * PACKETRATE;
  MACPDUSize=parentModule ()->parentModule ()->par ("MACPDUSize");
  RSCorrectionCapability=parentModule ()->parentModule ()->par("RSCorrectionCapability");
  RSBlockLength=parentModule ()->parentModule ()->par("RSBlockLength");
  //codedPHYSDUSize=(MACPDUSize*RSBlockLength)/(RSBlockLength-RSCorrectionCapability);
  codedPHYSDUSize=MACPDUSize;
  nodo.potenza = parentModule ()->parentModule ()->par ("riservapotenza");
  nodo.banda = PACKETRATE;//parentModule () ->parentModule ()->par ("PacketRate");
  connID = 0;
  PDUNumber = 0;
  indice = parentModule ()->index ();
  DATA_discarded_noise = 0;
  LC_discarded_noise = 0;
  LE_discarded_noise = 0;
  MissingDataSyncHeaderProb = 0;
  MissingCommonSyncHeaderProb = 0;
  DATAretransmitted=0;
  DEBUG = parentModule ()->parentModule ()->par ("debug");
  PRESERVINGINTERVAL = 10 * (double) (area_dim * sqrt(2)) / LIGHTSPEED + 2 * ((double) MAXDATASIZE / PACKETRATE);
  retransmit_rate =0;
  SlotDuration = (double) (MAXDATASIZE+PHYHeaderSize)/PACKETRATE + TRANSMITTER_RESET_TIME;
  RETRANSMIT_RATE_STEP = 3*SlotDuration;
  maxRetransmissionsPerPacket=parentModule ()->parentModule ()->par("maxTransmissionAttempts");
  ev << "Preserving interval= " << PRESERVINGINTERVAL << endl;
  ev << "Slot Duration=" << SlotDuration << endl;

  maxPower = STDPOWER * pow (rangetx, 2);
  finesim = false;

  xr = genk_uniform (1, 0, area_dim);
  yr = genk_uniform (1, 0, area_dim);

  x = (double) xr;
  y = (double) yr;

  database[indice].x = x;
  database[indice].y = y;
  
  
  //initialization of each node with the initial energy value
  //database[indice].NodeEnergy=InitialNodeEnergy; 
  NodeResidualEnergy[indice]=InitialNodeEnergy;
  NodePktEnergyCnt[indice]=0;
  //cout << "inizio" << NodeResidualEnergy[indice]<< endl;
  //cout << "initial energy of node "<< indice << " is "<< database[indice].NodeEnergy<< endl;

  inizializza_file_position (indice, esecuzione);
  print_coordinate (indice, esecuzione);

  px = areadisplay / area_dim * xr;
  py = areadisplay / area_dim * (area_dim - yr);
  mobilityupdate = 0;
  ev << "mobilityupdate " << mobilityupdate << ":";
  ev << "nodo " << indice << " ready:\n";
  ev << "x = " << x << endl;
  ev << "y = " << y << endl;
  ev << "mobilityupdate " << mobilityupdate << " eseguito;\n";

  routingtable = (struct locationinfo *) malloc (numnodi * sizeof (struct locationinfo));	//definizione  tavola routing  
  for (int i = 0; i < numnodi; i++)
    {
      //attivita[i] = 0;
      routingtable[i].valido = false;	//tavola routing viene invalidata
      routingtable[i].nexthop = -1;
      NPackets[i] = 0;
      TDATA[i] = TIMEOUTDATA;
      //      TDATA[i] = 0;
      ActiveFor[i] = true;
      Active[i] = true;
      //      TLC[i] = 0;
      TLC[i] = TIMEOUTLC;
      //      TActive[i] = 0;
      TActive[i] = TIMEOUTActive;
    }
  numeropacchetti.DATAinviati = 0;
  numeropacchetti.DATAricevuti = 0;
  numeropacchetti.DATAmissedSync = 0;
  numeropacchetti.DATALost = 0;
  numeropacchetti.DATACollided = 0;
  numeropacchetti.DATADropped = 0;

  numeropacchetti.LEinviati = 0;
  numeropacchetti.LEricevuti = 0;
  numeropacchetti.LEmissedSync = 0;
 
  numeropacchetti.LCinviati = 0;
  numeropacchetti.LCricevuti = 0;
  numeropacchetti.LCmissedSync = 0;
	        
  numeropacchetti.ACKDATAinviati = 0;
  numeropacchetti.ACKDATAricevuti = 0;
  numeropacchetti.ACKDATAmissedSync = 0;
  RecPackets = 0;
  AveragePEP = 0;
  AverageBEP = 0;
  NbInterferentiNodo=0;
  packetdelay = 0;
  cMessage *msg1 = new cMessage;	//inizializza generazione traffico 
  msg1->setKind (GENERATRAFFICO);
  msg1->setName ("GENERATRAFFICO");
  scheduleAt (simTime () + exponential(interarrivo, 1), msg1);

  cMessage *msg2 = new cMessage;	//inizializza generazione statistiche 
  msg2->setKind (AGGIORNASTA);
  msg2->setName ("AGGIORNASTA");
  scheduleAt (simTime (), msg2);

  cMessage *msg3 = new cMessage;	//aggiorna lista pacchetti 
  msg3->setKind (REFRESHLIST);
  msg3->setName ("REFRESHLIST");
  scheduleAt (simTime (), msg3);
  delaystat.setName ("delaystat");

  if (mobilita)			//caso mobilita
    {				//inizializza mobupdate con intervallo mobilityinterval

      vr = genk_intuniform (1, 0, vmax);
      dirr = genk_intuniform (1, 0, 360);

      v = (double) vr;
      dir = (double) dirr;

      cMessage *msg4 = new cMessage;
      msg4->setKind (GENERAMOBUPDATE);
      msg4->setName ("GENERAMOBUPDATE");
      msg4->setPriority (10);
      scheduleAt (simTime () + mobilityinterval, msg4);
    }
  pList = new FPL;
  //  connectionList = new connList;


}

// ##########################################
// Method name    : 
// Description    : descrive il comportamento del sistema quando si deve inviare o ricevere un pkt. Il comportamento  legato al tipo di pkt
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
void MediumAccessControl::activity ()
{

	cPar & x = parentModule ()->par ("x");	//passaggio parametri per variabile (&)
	cPar & y = parentModule ()->par ("y");	//assegnando variabili x,y,dir,v modifico parametri moduli relativi
	cPar & px = parentModule ()->par ("px");
	cPar & py = parentModule ()->par ("py");
	cPar & dir = parentModule ()->par ("direction");
	cPar & v = parentModule ()->par ("speed");
	//cPar & NodeEnergy = parentModule ()->par ("NodeEnergy");

	ofstream ratemanage;
	PDUNumber = 0;
	connID = 0;
	ev << "PacketRate: "<< PACKETRATE << endl;
	
	//shadow_term_802_15_4a = normal(0,1);
	while (!finesim)		//ciclo con durata tempo simulazione
	{

		//breakpoint("before receive");
		cMessage *msg = receive ();

		int kind = msg->kind ();
		ev << "Node [" << indice << "] : received message of kind " << kind << endl;
		
		//this flag is used for evaluating the pkt energy with the PacketEnergyEvaluation() method
		if (kind==SENDPACKET)
			SendingPacketFlag=true;
		else
			SendingPacketFlag=false;
		
		//      breakpoint("here");
		switch (kind)		//istruzione switch sul carattere tipomsg 
		{

			case GENERAMOBUPDATE:	//update posizione nodo secondo modello movimento
			{
				double theta = (double) dir / 360 * 2 * pi;

				ev << "mobilityupdate " << mobilityupdate << " : ";
				ev << "nodo " << indice << " start:\n";
				ev << "x = " << x << endl;
				ev << "y = " << y << endl;


				xr = (double) x + ((double) v * mobilityinterval * cos (theta));
				yr = (double) y + ((double) v * mobilityinterval * sin (theta));

				ev << "nodo " << indice << " intermediate values:\n";
				ev << "xr = " << xr << endl;
				ev << "yr = " << yr << endl;

				x = (double) xr;
				y = (double) yr;

				ev << "x = " << x << endl;
				ev << "y = " << y << endl;

				int rimbalzo = 0;

				bool dentrorete = ((double) x <= area_dim && (double) x >= 0 && (double) y <= area_dim && (double) y >= 0);


				while (!dentrorete && rimbalzo < 100)
				{
					if (xr > area_dim)
					{
						ev << " rimbalza x>area_dim\n";
						rimbalzo += 1;
						double deltax = fabs (xr - area_dim);
						x = (double) (area_dim - deltax);
					}


					if (xr < 0)
					{
						ev << " rimbalza x<0\n";
						rimbalzo += 1;
						double deltax = fabs (xr);
						x = (double) deltax;
					}

					if (yr > area_dim)
					{
						ev << " rimbalza y<area_dim\n";
						rimbalzo += 1;
						double deltay = fabs (yr - area_dim);
						y = area_dim - deltay;
					}


					if (yr < 0)
					{
						ev << " rimbalza y<0\n";
						rimbalzo += 1;
						double deltay = fabs (yr);
						y = (double) deltay;
					}
				dentrorete = ((double) x <= area_dim && (double) x >= 0 && (double) y <= area_dim && (double) y >= 0);
				}//end of "while (!dentrorete && rimbalzo < 100)"

				database[indice].x = x;
				database[indice].y = y;

				//print_coord_update (simTime (), indice);
				print_coordinate (indice, esecuzione);

				px = areadisplay / area_dim * (double) x;
				py = areadisplay / area_dim * (area_dim - (double) y);
				mobilityupdate += 1;

				ev << "mobilityupdate " << mobilityupdate << " : ";
				ev << "nodo " << indice << " ready:\n";
				ev << "x = " << x << endl;
				ev << "y = " << y << endl;
				ev << " mobilityupdate " << mobilityupdate << " eseguito;\n";


				double inerzia = uniform (0, 1);
				if (inerzia < correlazionemobilita)	//mobility model inertia: rallento modello mobilita': cond. con correlazionemobilita
				{
					dirr = genk_intuniform (1, 0, 360);
					vr = genk_intuniform (1, 0, vmax);
					dir = (double) dirr;
					v = (double) vr;
				}
				cMessage *msg2 = new cMessage;
				msg2->setKind (GENERAMOBUPDATE);
				msg2->setName ("GENERAMOBUPDATE");
				msg2->setPriority (10);
				scheduleAt (simTime () + mobilityinterval, msg2);
				delete msg;
			}//end of case case GENERAMOBUPDATE:
				break;

			case SENDPACKET:
			{
				int kind = (int) msg->par ("packetKind");
				

				switch (kind)
				{
					case DATA:
					{
						double packetDuration = (double) msg->length() / PACKETRATE;
						ev << "Node [" << indice << "] : sending DATA packet to " << msg->par("RXNodeID") << endl;
						//double txPower = evaluateTxPower (msg->par("RXNodeID")); //Commented out by Luca, 21/03/05;
						double txPower = Ptx;
						//cMessage *msg1 = new cMessage;
						
						//evaluation of the used energy for transmitting the next packet
						//cout << "node "<< indice << "sends a pkt, its residual energy was"<<database[indice].NodeEnergy;
						//cout<< NodeResidualEnergy << endl;
						//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodeResidualEnergy[indice]-=(double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodePktEnergyCnt[indice]++;
						
						
						//cout << " residual energy1 "<< database[indice].NodeEnergy<<" residual nrg 2 "<< NodeResidualEnergy[indice]<< endl;
				  
						msg->setName ("DATA");
						msg->setKind (DATA);
						msg->addPar ("txPower") = txPower;
						//msg1->addPar ("PDUNumber") = msg->par("PDUNumber");
				  
						double propagationDelay = sqrt (pow (database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow (database[indice].y - database[msg->par("RXNodeID")].y, 2)) / LIGHTSPEED;
						ev << "Propagation delay= " << propagationDelay << endl;
				  
						if (networkconnectivity[indice][ msg->par("RXNodeID")])
							sendDelayed (msg, packetDuration + propagationDelay, "outnode",  msg->par("RXNodeID") );	//19/09/03: Packet sent only if nodes are connected;
						else
						{
							ev << "Node [" << indice << "] disconnected between SENDPACKET generation and packet send to node [" << msg->par("RXNodeID") << "]" << endl;
							delete msg;
						}
						endTransmission[indice]=simTime()+packetDuration;
						transmittedPower[indice]=txPower;
						numeropacchetti.DATAinviati++;
						if(!PC_ENHANCED)
							packList->insertInFPLNoCheck (indice, msg->par("PDUNumber"), connID,
						
						simTime (), DATA, packetDuration, database[indice].x, database[indice].y, txPower);
					}//end of case "DATA"
						break;
					case ACKDATA:
					{
						
						//evaluation of the used energy for transmitting the next packet
						//cout << "node "<< indice << "sends a pkt, its residual energy was"<<database[indice].NodeEnergy; 
						//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodeResidualEnergy[indice]-= (double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodePktEnergyCnt[indice]++;
						//cout << " and now is "<< database[indice].NodeEnergy<< endl;
						
						double packetDuration = (double) msg->length() / PACKETRATE;
						//double txPower = evaluateTxPower (msg->par("RXNodeID"));
						double txPower = Ptx;
						//cMessage *msg1 = new cMessage;
						msg->setKind (ACKDATA);
						msg->setName ("ACKDATA");
						//msg1->addPar ("PDUNumber") = msg->par("PDUNumber");
						msg->addPar ("txPower") = txPower;
						double propagationDelay =
							sqrt (pow (database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow (database[indice].y - database[msg->par("RXNodeID")].y, 2)) / LIGHTSPEED;

						if (networkconnectivity[indice][msg->par("RXNodeID")])
							sendDelayed (msg, packetDuration + propagationDelay, "outnode", msg->par("RXNodeID"));
						else
						{
							ev << "Node [" << indice << "] disconnected between SENDPACKET generation and packet send to node [" << msg->par("RXNodeID") << "]" <<  endl;
							delete msg;
						}
						nodo.potenza -= txPower;
						numeropacchetti.ACKDATAinviati++;
						if(!PC_ENHANCED)
							packList->insertInFPLNoCheck (indice, msg->par("PDUNumber"), connID,
						simTime (), ACKDATA, packetDuration, database[indice].x, database[indice].y, txPower);
					}//end of case ACKDATA
						break;
					//LE packet
					case LE:
					{
						//evaluation of the used energy for transmitting the next packet
						//cout << "node "<< indice << "sends a pkt, its residual energy was"<<database[indice].NodeEnergy; 
						//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodeResidualEnergy[indice]-= (double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodePktEnergyCnt[indice]++;
						//cout << " and now is "<< database[indice].NodeEnergy<< endl;
						
						//breakpoint("here: received SENDPACKET_LE packet");
						double packetDuration = (double) msg->length() / PACKETRATE;
						//double txPower = evaluateTxPower (msg->par("RXNodeID"));
						double txPower = Ptx;
						msg->setKind(LE);
						msg->setName("LE");
						msg->addPar ("txPower") = txPower;
						double propagationDelay =
							sqrt (pow (database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow (database[indice].y - database[msg->par("RXNodeID")].y, 2)) / LIGHTSPEED;

						if(networkconnectivity[indice][msg->par("RXNodeID")])
							sendDelayed (msg, packetDuration + propagationDelay, "outnode", msg->par("RXNodeID"));
						else
						{
							ev <<"Node [" << indice << "]:  networkconnectivity with RXNode = " << networkconnectivity[indice][msg->par("RXNodeID")] << endl;

							ev << "Node [" << indice << "] disconnected between SENDPACKET generation and packet send to node [" << msg->par("RXNodeID") << "]" <<  endl;
							//breakpoint("here: received SENDPACKET_LE packet but disconnected");
							delete msg ;
						}
						nodo.potenza -= txPower;
						numeropacchetti.LEinviati++;
						if(!PC_ENHANCED)		  
							packList->insertInFPLNoCheck (indice, msg->par("PDUNumber"), connID,
						
						simTime (), LE, packetDuration, database[indice].x, database[indice].y, txPower);
						TLC[msg->par("RXNodeID")] = simTime() + TIMEOUTLC;
						//breakpoint("here: sent LE packet, TLC set");
					}//end of case LE
						break;
					//LC packet
					case LC:
					{
						
						//evaluation of the used energy for transmitting the next packet
						//cout << "node "<< indice << "sends a pkt, its residual energy was"<<database[indice].NodeEnergy; 
						//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodeResidualEnergy[indice]-= (double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodePktEnergyCnt[indice]++;
						//cout << " and now is "<< database[indice].NodeEnergy<< endl;
						
						double packetDuration = (double) msg->length() / PACKETRATE;
						//double txPower = evaluateTxPower (msg->par("RXNodeID"));
						double txPower = Ptx;
						msg->setKind(LC);
						msg->setName("LC");
						msg->addPar ("txPower") = txPower;
						double propagationDelay =
							sqrt (pow (database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow (database[indice].y - database[msg->par("RXNodeID")].y, 2)) / LIGHTSPEED;

						if (networkconnectivity[indice][msg->par("RXNodeID")])
							sendDelayed (msg, packetDuration + propagationDelay, "outnode", msg->par("RXNodeID"));
						else
						{
							ev << "Node [" << indice << "] disconnected between SENDPACKET generation and packet send to node [" << msg->par("RXNodeID") << "]" <<  endl;
							delete msg ;
						}
						nodo.potenza -= txPower;
						numeropacchetti.LCinviati++;
						if(!PC_ENHANCED)
							packList->insertInFPLNoCheck (indice, msg->par("PDUNumber"), connID,
						simTime (), LC, packetDuration, database[indice].x, database[indice].y, txPower);
					}
						break;
					case DISTANCE_REQUEST:
					{
						//evaluation of the used energy for transmitting the next packet
						//cout << "node "<< indice << "sends a pkt, its residual energy was"<<database[indice].NodeEnergy; 
						//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodeResidualEnergy[indice]-= (double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodePktEnergyCnt[indice]++;
						//cout << " and now is "<< database[indice].NodeEnergy<< endl;
						//cout << " residual energy1 "<< database[indice].NodeEnergy<<" residual nrg 2 "<< NodeResidualEnergy[indice]<< endl;

						double packetDuration = (double) msg->length () / PACKETRATE;
						int targetNode = msg->par ("destinationNode");
						ev << "Node ["<<indice<<"]: sending a DISTANCE_REQUEST to "<< targetNode << endl;
						breakpoint("Sending DISTANCE_REQUEST");
						//double txPower = evaluateTxPower (targetNode);
						double txPower = Ptx;
						//cMessage *msg1 = new cMessage;
						msg->setKind (DISTANCE_REQUEST);
						msg->setTimestamp ();
						msg->setName ("DISTANCE_REQ");
						//msg->addPar ("PDUNumber") = PDUNumber;
						msg->addPar ("txPower") = txPower;
						double propagationDelay =
							sqrt (pow (database[indice].x - database[targetNode].x, 2) + pow (database[indice].y - database[targetNode].y, 2)) / LIGHTSPEED;
				  
						if(networkconnectivity[indice][targetNode])
							sendDelayed (msg, packetDuration + propagationDelay, "outnode", targetNode);
						else
						{
							ev <<  "Node [" << indice << "] disconnected between SENDPACKET generation and packet send to node [" << targetNode << "]"<< endl;
							delete msg;
						}
						nodo.potenza -= txPower;
						//                      numeropacchetti.ACKDATAinviati++;
						
						if(!PC_ENHANCED)
							packList->insertInFPLNoCheck (indice, msg->par("PDUNumber"), connID,
						simTime (), DISTANCE_REQUEST, packetDuration, database[indice].x, database[indice].y, txPower);
						//PDUNumber++;
						endTransmission[indice]=simTime()+packetDuration;
						transmittedPower[indice]=txPower;
					}
						break;

					case DISTANCE_REPLY:
					{
						
						//evaluation of the used energy for transmitting the next packet
						//cout << "node "<< indice << "sends a pkt, its residual energy was"<<database[indice].NodeEnergy; 
						//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodeResidualEnergy[indice]-= (double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[msg->par("RXNodeID")].x, 2) + pow(database[indice].y - database[msg->par("RXNodeID")].y, 2)),  (int) msg->length(), SendingPacketFlag);
						NodePktEnergyCnt[indice]++;
						//cout << " and now is "<< database[indice].NodeEnergy<< endl;
						
						double packetDuration = (double) msg->length () / PACKETRATE;
						int targetNode = msg->par ("destinationNode");
						//double txPower = evaluateTxPower (targetNode);
						double txPower = Ptx;
						//cMessage *msg1 = new cMessage;
						msg->setKind (DISTANCE_REPLY);
						msg->setTimestamp ();
						msg->setName ("DISTANCE_REP");
						//msg1->addPar ("PDUNumber") = msg->par("PDUNumber");
						msg->addPar ("txPower") = txPower;
						double propagationDelay =
							sqrt (pow (database[indice].x - database[targetNode].x, 2) + pow (database[indice].y - database[targetNode].y, 2)) / LIGHTSPEED;
						if(networkconnectivity[indice][targetNode])
							sendDelayed (msg, packetDuration + propagationDelay, "outnode", targetNode);
						else
						{
							ev << "Disconnected between SENDPACKET generation and packet send... " << endl;
							delete msg;
						}
						nodo.potenza -= txPower;
						//                      numeropacchetti.ACKDATAinviati++;
						if(!PC_ENHANCED)
							packList->insertInFPLNoCheck (indice, msg->par("PDUNumber"), connID,
								simTime (), DISTANCE_REPLY, packetDuration, database[indice].x, database[indice].y, txPower);
						//PDUNumber++;
						endTransmission[indice]=simTime()+packetDuration;
						transmittedPower[indice]=txPower;
					}//end of case case DISTANCE_REPLY:
						break;
				}//end of switch (kind) (inner switch)
			}//end of case SENDPACKET:
				break;

			case GENERATRAFFICO:
			{
			if (simTime () <= TOsim)
			{
				int NPDU, RXNodeID;
				bool connesso = false;
				for (int j = 0; j < numnodi; j++)
					connesso = connesso || networkconnectivity[indice][j];
				//i due nodi sono in visibilit
				if (connesso)
				{
					ev << "Node [" << indice << "] manages trafficGen" << endl;
					trafficGen(&NPDU, &RXNodeID);
					//while (!networkconnectivity[indice][RXNodeID])
					//   trafficGen(&NPDU, &RXNodeID);
					ev << "Node [" << indice << "]: NPDU = " << NPDU << ", RXNodeID = " << RXNodeID << endl;
					
					for (int i=0; i < NPDU; i++)
						TXQueue.push_back(RXNodeID);
					
					NPackets[RXNodeID] += NPDU - 1;

					// TLC[RXNodeID] = simTime() + TIMEOUTLC; //Hack! We set this value here just to help the serveTXQueue function to undestand that a LE packet to RXNodeID must be sent: TLC is really set when the LE packet is effectively sent.
					if(TActive[RXNodeID] < simTime())
						Active[RXNodeID] = false;
					
					msg->addPar("TXTime")=simTime();//added
					msg->addPar("TXAttempt")=1;
					serveTXQueue(msg,RXNodeID);
				}//end of "if connesso"
			
				double nextinterarrivo = exponential(interarrivo, 1);
				while (nextinterarrivo == 0)
					nextinterarrivo = exponential(interarrivo, 1);
				
				if(DEBUG)
				{
					ofstream inter;
					inter.open("Interarrival_Time.txt", ios::app);
					inter << nextinterarrivo << endl;
					inter.close();
				}
				cMessage *msg2 = new cMessage;
			
				msg2->setKind (GENERATRAFFICO);
				msg2->setName ("GENERATRAFFICO");
				scheduleAt (simTime () + nextinterarrivo, msg2);
			}//end of "if (simTime () <= TOsim)"
			else
			{
				ev << "nodo " << indice << " ha finito\n";
				ev << simTime () << endl;

				cMessage *fine = new cMessage;
				fine->setKind (FINE);
				cModule *connectivity = parentModule ()->parentModule ()->submodule ("connectivity");
				cGate *inconi = connectivity->gate ("inconnectivity", indice);
				sendDirect (fine, 10, inconi);
			}
			delete msg;
		}//end of case GENERATRAFFICO:
			break;
		
				
		//RECEIVED PACKET MANAGEMENT		
		case DATA:		//gestione pacchetto data
		{
						
			int id_MAC_TX = (int) msg->senderModuleId ();
			cModule *MAC_TX_node = simulation.module (id_MAC_TX);
			int txNodeID = (MAC_TX_node->parentModule ())->index ();
			
			//evaluation of the used energy for the reception of this packet (UWB case)
			//Energy[indice] = (double) Energy[indice] - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			NodeResidualEnergy[indice]-=(double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			NodePktEnergyCnt[indice]++;

			if (ActiveFor[txNodeID] && simTime() < TDATA[txNodeID])
			{
				double value = genk_uniform (1, 0, 1);

				if (value < MissingDataSyncHeaderProb)
				{
					ev << "Node [" << indice << "],  missed  SyncHeader for packet of kind " << DATA << " (DATA)" << endl;
					numeropacchetti.DATAmissedSync++;
				}
				else
				{
					int id_MAC_TX = (int) msg->senderModuleId ();
					cModule *MAC_TX_node = simulation.module (id_MAC_TX);

					int txNodeID = (MAC_TX_node->parentModule ())->index ();
					ev << "Node [" << indice << "],  received packet of kind " << DATA << " (DATA)" << " from node[" << txNodeID << endl;

					double packetDuration = (double) msg->length() / PACKETRATE;
					bool pStatus = checkCorrectReceiving (DATA, msg->length() - PHYHeaderSize, (int) msg->par ("PDUNumber"), simTime (), msg->par ("txPower"), txNodeID, msg);

					if ((pStatus) && (((int) (msg->par("RXNodeID")) ) == indice))
					{
						numeropacchetti.DATAricevuti++;
						packetdelay += simTime() - (double) msg->par("TXTime");
						ev<<"Node ["<<indice<<"]: packet delay = "<<  simTime() - (double) msg->par("TXTime") << ", packet duration = " << packetDuration << endl; 
						cMessage *ack = (cMessage *) msg->dup();
						ack->setKind (ACKDATA); 
						ack->setName("ACK");
						ack->par("TXNodeID")=indice;
						ack->par("RXNodeID")=txNodeID;
						//ack->addPar("RXNodeID")=indice;
						cGate *ok= MAC_TX_node->gate("innode", indice);
						double packetDuration = (double) ACKDATASIZE / PACKETRATE;
						//sendDirect(ack, ok);
						send(ack,"outnode",txNodeID);
						retransmit_rate=0;			

						if (!(NeighbourMatrix[indice][txNodeID].distanceValidity) || simTime () - NeighbourMatrix[indice][txNodeID].time > DISTANCE_TIMEOUT)
						{
							cMessage *distupdate = (cMessage *) msg->dup ();
							ev << "TXNodeID: " << txNodeID << endl;
							//distupdate->addPar ("txNodeID") = txNodeID;
							send (distupdate, "toLOC");
						}
					}
					else//added the else
					{
						cMessage *nack = (cMessage *) msg->dup();
						nack->setKind (NACK); 
						nack->setName("NACK");
						nack->par("TXNodeID")=indice;
						nack->par("RXNodeID")=txNodeID;
						cGate * error= MAC_TX_node->gate("innode", indice);
						double packetDuration = (double) ACKDATASIZE / PACKETRATE;
						//sendDirect(nack, error);
						send(nack, "outnode", txNodeID);
						numeropacchetti.DATACollided++;
						breakpoint("Wrong DATA Packet");
					}

					if(simTime() - NeighbourMatrix[indice][txNodeID].lastPacketTime > LASTPACKET_TIMEOUT)
						NeighbourMatrix[indice][txNodeID].receivedPacketsMemory = 0;
					
					NeighbourMatrix[indice][txNodeID].receivedPacketsMemory++;
					NeighbourMatrix[indice][txNodeID].lastPacketTime = simTime();
					TDATA[txNodeID] = simTime() + TIMEOUTDATA;
				}//end of else

			}//end of if (ActiveFor[txNodeID] && simTime() < TDATA[txNodeID])
			else
			{
				ActiveFor[txNodeID] = false;
				ev << "Node [" << indice << "],  did not listen to SyncHeader for packet of kind " << DATA << " (DATA)" << endl;
				numeropacchetti.DATALost++;
			}
			delete msg;
		}
		break;
			
		case ACKDATA:
		{

			int id_MAC_TX = (int) msg->senderModuleId ();
			cModule *MAC_TX_node = simulation.module (id_MAC_TX);
			int txNodeID = (MAC_TX_node->parentModule ())->index ();
			
			//evaluation of the used energy for the reception of this packet (UWB case)
			//Energy[indice] = (double) Energy[indice] - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			NodeResidualEnergy[indice]-=(double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			NodePktEnergyCnt[indice]++;
			
			
			double value = genk_uniform (1, 0, 1);
			if (value < MissingCommonSyncHeaderProb)
			  {
			ev << "Node [" << indice << "],  missed  SyncHeader for packet of kind " << ACKDATA << " (ACKDATA)" << endl;
			numeropacchetti.ACKDATAmissedSync++;
			  }
			else
			  {	
			//Hack: ACKDATA packets are always correct... 
			//bool pStatus = checkCorrectReceiving (ACKDATA, msg->length() - PHYHeaderSize, (int) msg->par ("PDUNumber"), simTime (), msg->par ("txPower"), txNodeID, msg);
			//if (pStatus && (((int) msg->par("RXNodeID")) == indice))
			//{
				ev << "Node [" << indice << "]  received packet of kind " << ACKDATA << " (ACKDATA)" << endl;
				numeropacchetti.ACKDATAricevuti++;
				if (!(NeighbourMatrix[indice][txNodeID].distanceValidity) || simTime () - NeighbourMatrix[indice][txNodeID].time > DISTANCE_TIMEOUT)
				  {
				cMessage *distupdate = (cMessage *) msg->dup ();
				ev << "TXNodeID: " << txNodeID << endl;
				//distupdate->addPar ("txNodeID") = txNodeID;
				send (distupdate, "toLOC");
				  }
				if(simTime() - NeighbourMatrix[indice][txNodeID].lastPacketTime > LASTPACKET_TIMEOUT)
				  NeighbourMatrix[indice][txNodeID].receivedPacketsMemory = 0;
				NeighbourMatrix[indice][txNodeID].receivedPacketsMemory++;
				NeighbourMatrix[indice][txNodeID].lastPacketTime = simTime();
				//MISSING: ACKDATA MANAGEMENT (CHECK OF THE CORRESPONDING DATA PACKET); NOT ENOUGH TIME
				//}
			  }
			delete msg;
		  }
		  break;
		  case LE:
			{
				
			  int id_MAC_TX = (int) msg->senderModuleId ();
			  cModule *MAC_TX_node = simulation.module (id_MAC_TX);
			  int txNodeID = (MAC_TX_node->parentModule ())->index ();
			  
			  //evaluation of the used energy for the reception of this packet (UWB case)
			 // Energy[indice] = (double) Energy[indice] - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			  //database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			  NodeResidualEnergy[indice]-=(double)  PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			  NodePktEnergyCnt[indice]++;
			  
			  double value = genk_uniform (1, 0, 1);

			  if (value < MissingCommonSyncHeaderProb)
			{
			  ev << "Node [" << indice << "],  missed  SyncHeader for packet of kind " << LE << " (LE)" << endl;
			  numeropacchetti.LEmissedSync++;
			}
			  else
			{
			  bool pStatus = checkCorrectReceiving (LE, msg->length() - PHYHeaderSize, (int) msg->par ("PDUNumber"), simTime (), msg->par ("txPower"), txNodeID, msg);
			  if (pStatus && (((int)msg->par("RXNodeID")) == indice))
				{
				  ev << "Node [" << indice << "]  received packet of kind " << LE << " (LE)" << endl;
				  numeropacchetti.LEricevuti++;
				  if (!(NeighbourMatrix[indice][txNodeID].distanceValidity) || simTime () - NeighbourMatrix[indice][txNodeID].time > DISTANCE_TIMEOUT)
				{
				  cMessage *distupdate = (cMessage *) msg->dup ();
				  ev << "TXNodeID: " << txNodeID << endl;
				  //distupdate->addPar ("TXNodeID") = txNodeID;
				  send (distupdate, "toLOC");
				}
				  if(simTime() - NeighbourMatrix[indice][txNodeID].lastPacketTime > LASTPACKET_TIMEOUT)
				NeighbourMatrix[indice][txNodeID].receivedPacketsMemory = 0;
				  NeighbourMatrix[indice][txNodeID].receivedPacketsMemory++;
				  NeighbourMatrix[indice][txNodeID].lastPacketTime = simTime();
				  
				  packetScheduling (LCSIZE, 0, msg->par ("TXNodeID"), &PDUNumber, LC, simTime (), msg);		      
				  ActiveFor[ msg->par ("TXNodeID")]=true;
				  TDATA[msg->par("TXNodeID")] = simTime() + TIMEOUTDATA; 
				}
			}
			  delete msg;
			}
			break;

		  case LC:
			{

			  int id_MAC_TX = (int) msg->senderModuleId ();
			  cModule *MAC_TX_node = simulation.module (id_MAC_TX);
			  int txNodeID = (MAC_TX_node->parentModule ())->index ();
			  
			  //evaluation of the used energy for the reception of this packet (UWB case)
			 // Energy[indice] = (double) Energy[indice] - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			  //database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			  NodeResidualEnergy[indice]-=(double)  PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			  NodePktEnergyCnt[indice]++;
			  
			  double value = genk_uniform (1, 0, 1);
			  if (value < MissingCommonSyncHeaderProb)
			{
			  ev << "Node [" << indice << "],  missed  SyncHeader for packet of kind " << LC << " (LC)" << endl;
			  numeropacchetti.LCmissedSync++;
			}
			  else
			{	    
			  //breakpoint("Received LC, checking if it is correct");
			  bool pStatus = checkCorrectReceiving (LC, msg->length()  - PHYHeaderSize, (int) msg->par ("PDUNumber"), simTime (), msg->par ("txPower"), txNodeID, msg);
			  if (pStatus && (((int)msg->par("RXNodeID")) == indice))
				{
				  ev << "Node [" << indice << "]  received packet of kind " << LC << " (LC)" << endl;
				  numeropacchetti.LCricevuti++;
				  breakpoint("Received LC, it is correct");
				  if (!(NeighbourMatrix[indice][txNodeID].distanceValidity) || simTime () - NeighbourMatrix[indice][txNodeID].time > DISTANCE_TIMEOUT)
				{
				  cMessage *distupdate = (cMessage *) msg->dup ();
				  ev << "TXNodeID: " << txNodeID << endl;
				  //distupdate->addPar ("tXNodeID") = txNodeID;
				  send (distupdate, "toLOC");
				}
				  if(simTime() - NeighbourMatrix[indice][txNodeID].lastPacketTime > LASTPACKET_TIMEOUT)
				NeighbourMatrix[indice][txNodeID].receivedPacketsMemory = 0;
				  NeighbourMatrix[indice][txNodeID].receivedPacketsMemory++;
				  NeighbourMatrix[indice][txNodeID].lastPacketTime = simTime();
				  Active[msg->par("TXNodeID")] = true;
				  TActive[msg->par("TXNodeID")] = simTime() + TIMEOUTActive;
				  //breakpoint("Received LC, invoking serveTXQueue");
				  serveTXQueue(msg);
				}
			}
			  delete msg;
			  //breakpoint("Handled LC");
			}
			break;

		case DISTANCE_REQUEST:
		  {

			if (msg->arrivedOn ("fromLOC"))
			  {
			ev << "Node_[" << indice << "].MAC: DISTANCE_REQUEST received by Node_[" << indice << "].LOC" << endl;
			ev <<  "Node [" << indice << "], scheduling packet with ID " << PDUNumber << endl;
			packetScheduling (msg->length (), 0, msg->par ("destinationNode"), &PDUNumber, DISTANCE_REQUEST, simTime (), msg);
			  }
			else
			  {
			int id_MAC_TX = (int) msg->senderModuleId ();
			cModule *MAC_TX_node = simulation.module (id_MAC_TX);
			int txNodeID = (MAC_TX_node->parentModule ())->index ();
			
			//evaluation of the used energy for the reception of this packet (UWB case)
			//Energy[indice] = (double) Energy[indice] - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			NodeResidualEnergy[indice]-=(double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			NodePktEnergyCnt[indice]++;
			
			
			//breakpoint("Received DISTANCE_REQ from the channel");
			ev << "Node [" << indice << "] : received DISTANCE_REQ from [" << txNodeID << "]" << endl;
			bool pStatus = checkCorrectReceiving (DISTANCE_REQUEST, msg->length() - PHYHeaderSize, (int) msg->par ("PDUNumber"), simTime (), msg->par ("txPower"), txNodeID, msg);
			if(pStatus){
			  if(simTime() - NeighbourMatrix[indice][txNodeID].lastPacketTime > LASTPACKET_TIMEOUT)
				NeighbourMatrix[indice][txNodeID].receivedPacketsMemory = 0;
			  NeighbourMatrix[indice][txNodeID].receivedPacketsMemory++;
			  NeighbourMatrix[indice][txNodeID].lastPacketTime = simTime();
			  cMessage *distupdate = (cMessage *) msg->dup ();
			  distupdate->addPar ("txNodeID") = txNodeID;
			  send (distupdate, "toLOC");
			}
			else
			  {
				ev << "Node [" << indice << "]: packet DISTANCE_REQUEST discarded\n";
			  }

			  }
			delete msg;
		  }
		  break;

		case DISTANCE_REPLY:
		  {

			if (msg->arrivedOn ("fromLOC"))
			  {
			ev << "Node_[" << indice << "].MAC: DISTANCE_REPLY received by Node_[" << indice << "].LOC" << endl;
			ev <<  "Node [" << indice << "], scheduling packet with ID " << PDUNumber << endl;
			packetScheduling ( msg->length (), 0, msg->par ("destinationNode"), &PDUNumber, DISTANCE_REPLY, simTime (), msg);

			  }
			else
			  {
			int id_MAC_TX = (int) msg->senderModuleId ();
			cModule *MAC_TX_node = simulation.module (id_MAC_TX);
			int txNodeID = (MAC_TX_node->parentModule ())->index ();
			
			//evaluation of the used energy for the reception of this packet (UWB case)
			//Energy[indice] = (double) Energy[indice] - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			//database[indice].NodeEnergy = (double) database[indice].NodeEnergy - PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			NodeResidualEnergy[indice]-=(double) PacketEnergyEvaluation(sqrt(pow(database[indice].x - database[txNodeID].x, 2) + pow(database[indice].y - database[txNodeID].y, 2)),  (int) msg->length(), SendingPacketFlag);
			NodePktEnergyCnt[indice]++;
			
			bool pStatus = checkCorrectReceiving (DISTANCE_REPLY, msg->length() - PHYHeaderSize, (int) msg->par ("PDUNumber"), simTime (), msg->par ("txPower"), txNodeID, msg);
			if(pStatus)
			  {
				if(simTime() - NeighbourMatrix[indice][txNodeID].lastPacketTime > LASTPACKET_TIMEOUT)
				  NeighbourMatrix[indice][txNodeID].receivedPacketsMemory = 0;
				NeighbourMatrix[indice][txNodeID].receivedPacketsMemory++;
				NeighbourMatrix[indice][txNodeID].lastPacketTime = simTime();
				cMessage *distupdate = (cMessage *) msg->dup ();
				ev << "Node_[" << indice << "].MAC: DISTANCE_REPLY received by Node_[" << txNodeID << "]" << endl;
			  //distupdate->addPar ("txNodeID") = txNodeID;
				send (distupdate, "toLOC");
			  }
			else
			  {
				ev << "Node [" << indice << "]: DISTANCE_REPLY packet discarded\n";
			  }

			  }
			delete msg;
		  }
		  break;
		  case AGGIORNASTA:
			{
			  int DATAinviati = numeropacchetti.DATAinviati;
			  int DATAricevuti = numeropacchetti.DATAricevuti;
			  int DATAmissedSync = numeropacchetti.DATAmissedSync;
			  int DATALost = numeropacchetti.DATALost;
			  int DATACollided = numeropacchetti.DATACollided;
			  int DATADropped = numeropacchetti.DATADropped;
			  
			  int LEinviati = numeropacchetti.LEinviati;
			  int LEricevuti = numeropacchetti.LEricevuti;
			  int LEmissedSync = numeropacchetti.LEmissedSync;
			  
			  int LCinviati = numeropacchetti.LCinviati;
			  int LCricevuti = numeropacchetti.LCricevuti;
			  int LCmissedSync = numeropacchetti.LCmissedSync;
			  
			  int ACKDATAinviati = numeropacchetti.ACKDATAinviati;
			  int ACKDATAricevuti = numeropacchetti.ACKDATAricevuti;
			  int ACKDATAmissedSync = numeropacchetti.ACKDATAmissedSync;

			  pacchettiDATAinviati[indice] = DATAinviati;
			  pacchettiDATAricevuti[indice] = DATAricevuti;
			  pacchettiDATAmissedSync[indice] = DATAmissedSync;
			  pacchettiDATALost[indice] = DATALost;
			  pacchettiDATAcollided[indice] = DATACollided;
			  pacchettiDATADropped[indice]=DATADropped;
			  pacchettiACKDATAinviati[indice] = ACKDATAinviati;
			  pacchettiACKDATAricevuti[indice] = ACKDATAricevuti;
			  pacchettiACKDATAmissedSync[indice] = ACKDATAmissedSync;

			  pacchettiLEinviati[indice] = LEinviati;
			  pacchettiLEricevuti[indice] = LEricevuti;
			  pacchettiLEmissedSync[indice] = LEmissedSync;

			  pacchettiLCinviati[indice] = LCinviati;
			  pacchettiLCricevuti[indice] = LCricevuti;
			  pacchettiLCmissedSync[indice] = LCmissedSync;

			  PEP[indice] = AveragePEP / RecPackets;
			  BEP[indice] = AverageBEP / RecPackets;
			  //if (indice==1||indice==2)
			  //cout <<"indice "<< indice << " average ber " << AverageBEP << "diviso per rec packets "<<  RecPackets<<endl;
			  
			  NbInterferenti[indice]=NbInterferentiNodo / RecPackets;
			  //if (indice==1||indice==2)
				//  cout <<"indice "<< indice << " interferenti " << NbInterferentiNodo <<endl;
			  
			  NodesResidualEnergyVector[indice]=NodeResidualEnergy[indice];
			  TotPktEnergyCntVector[indice]=NodePktEnergyCnt[indice];
			  
			  averagedelay[indice] = packetdelay/DATAricevuti;
			  ev<<"Node["<<indice<<"]: DATAricevuti ="<< DATAricevuti<<",DATAritrasmessi ="<<DATAretransmitted<<",  average delay = "<< packetdelay/DATAricevuti<<endl;
			  double mdel = (double) delaystat.mean ();
			  double vdel = (double) delaystat.stddev ();
			  delayrtt[indice][0] = mdel;
			  delayrtt[indice][1] = vdel;
			  
			  cMessage *msg1 = new cMessage;
			  msg1->setKind (AGGIORNASTA);
			  msg1->setName ("AGGIORNASTA");
			  scheduleAt (simTime () + STATINTERVAL, msg1);
			  delete msg;
			}
			break;

		case REFRESHLIST:
		  {
			pList->refreshList (simTime (), PRESERVINGINTERVAL);
			cMessage *msg3 = new cMessage;
			msg3->setKind (REFRESHLIST);
			msg3->setName ("REFRESHLIST");
			scheduleAt (simTime () + REFRESHINGINTERVAL, msg3);
			delete msg;
		  }
		  break;

		case FINE:
		  {

			ofstream fp;
			fp.open ("global.sca", ios::app);
			double mdel = delaystat.mean ();
			double stddel = delaystat.stddev ();
			fp << "PARZIALE nodo [" << indice << "]" << endl; 
			fp << "Pacchetti scartati per rumore: LE = " << LE_discarded_noise << ", LC = " << LC_discarded_noise << ", DATA = " << DATA_discarded_noise << endl; 
			fp << "valore medio delayrtt: " << mdel << " deviazione standard delayrtt: ";
			fp << stddel << endl;
			fp.close ();
			ev << "aggiornamento file output.sca eseguito\n";
			cMessage *endloc = (cMessage *) msg->dup ();
			endloc->addPar("TXNodeID") = indice;
			send (endloc, "toLOC");
			finesim = true;
			free (routingtable);
			delete msg;
		  }
		  break;
		  //added from here
		case NACK:
		  {
			int attemptsNumber= (int) msg->par("TXAttempt");
			
			if(attemptsNumber<maxRetransmissionsPerPacket)
			  {
			double nextretransmit;
			if(BEB)
			  {  
				retransmit_rate = pow(RETRANSMIT_RATE_STEP,attemptsNumber);
				while (nextretransmit == 0)
				  nextretransmit = exponential(retransmit_rate, 1);
			  }
			else
			  nextretransmit =0;
			if(DEBUG)
			  {
				ofstream inter;
				inter.open("Interarrival_Time.txt", ios::app);
				inter << nextretransmit << endl;
				inter.close();
			  }
			
			cMessage * msg2 =  (cMessage *) msg->dup();
			msg2->setKind (RETRANSMIT);
			msg2->setName ("RETRANSMIT");
			msg2->par("TXAttempt")=attemptsNumber+1;
			scheduleAt (simTime () + nextretransmit, msg2);
			  }
			else
			  {
			numeropacchetti.DATACollided++;
			  }
			delete msg;
			  
		  } 
		  break;
		  //this is a send case  
		case RETRANSMIT:
		  { 
			int NPDU, RXNodeID;
			ev << "Node [" << indice << "] manages re-transmission" << endl;
			RXNodeID=msg->par("TXNodeID");
			
			
			NPDU=1;
			DATAretransmitted++;
			ev << "Node [" << indice << "]: NPDU = " << NPDU << ", RXNodeID = " << RXNodeID << endl;
			for (int i=0; i < NPDU; i++)
			  TXQueue.push_back(RXNodeID);
			NPackets[RXNodeID] += NPDU - 1;
			if(TActive[RXNodeID] < simTime())
			  Active[RXNodeID] = false;
			serveTXQueue(msg,RXNodeID);
			delete msg;
		  }
		  break;
    }
    }  
}				//snapshot(this);

// ##########################################
// Method name    : 
// Description    : 
// Arguments      : 
// Called by      : 
// Returned value : 
// ##########################################
void MediumAccessControl::finish ()
{
  ev << "nodo " << indice << " ha finito\n";
}
